/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
 * this module stores a user_struct which can be globally accessed (by root)
 * across process boundaries, in order to obtain info about users currently
 * accessing * samba. 
 *
 * the database key is the process number (pid) of the creater of the
 * structure plus the vuid - SMB virtual user id.  NORMALLY, this would
 * be smbd pid and an smbd-generated vuid.
 *
 */

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "others.h"


static TDB_CONTEXT *tdb = NULL;


static BOOL vuid_init_db(void)
{
	if (tdb)
		return True;

	tdb = tdb_open_log(NULL, 0, TDB_INTERNAL, O_RDWR | O_CREAT, 0600);

	if (tdb == NULL)
	{
		DEBUG(0,("vuid_init_db: failed\n"));
		return False;
	}
	
	DEBUG(10,("vuid_init_db: opened\n"));

	return True;
}

BOOL tdb_delete_vuid( uint16 vuid)
{
	prs_struct key;

	if (tdb == NULL)
	{
		if (!vuid_init_db())
		{
			return False;
		}
	}

	DEBUG(10,("delete user %x\n", vuid));

	prs_init(&key, MARSHALL);
	if (!_prs_uint16("vuid", &key, 0, &vuid))
	{
		return False;
	}

	prs_tdb_delete(tdb, &key);

	prs_free_data(&key);

	return True;
}

BOOL tdb_lookup_vuid( uint16 vuid, user_struct **usr)
{
	prs_struct key;
	prs_struct data;

	if (usr != NULL)
	{
		(*usr) = (user_struct *)malloc(sizeof(**usr));
		if ((*usr) == NULL)
		{
			return False;
		}
		ZERO_STRUCTP((*usr));
	}
	if (tdb == NULL)
	{
		if (!vuid_init_db())
		{
			safe_free((*usr));
			return False;
		}
	}

	DEBUG(10,("lookup user %x\n", vuid));

	prs_init(&key, MARSHALL);
	if (!_prs_uint16("vuid", &key, 0, &vuid))
	{
		prs_free_data(&key);
		safe_free((*usr));
		return False;
	}

	prs_tdb_fetch(tdb, &key, &data);

	if (usr != NULL)
	{
		if (!vuid_io_user_struct("usr", (*usr), &data, 100))
		{
			prs_free_data(&key);
			prs_free_data(&data);
			safe_free((*usr));
			return False;
		}
	}

	prs_free_data(&key);
	prs_free_data(&data);

	return True;
}

BOOL tdb_store_vuid(uint16 vuid, user_struct *usr, BOOL overwrite)
{
	prs_struct key;
	prs_struct data;
	BOOL ret = True;

	if (tdb == NULL)
	{
		if (!vuid_init_db())
		{
			return False;
		}
	}

	DEBUG(10,("storing user %x\n", vuid));

	prs_init(&key, MARSHALL);
	prs_init(&data, MARSHALL);

	if (!_prs_uint16("vuid", &key, 0, &vuid) ||
	    !vuid_io_user_struct("usr", usr, &data, 100))
	{
		DEBUG(0, ("tdb_store_vuid: marshalling error\n"));
		ret = False;
	}
	if (ret && (prs_tdb_store(tdb, (overwrite ? TDB_REPLACE : TDB_INSERT),
				  &key, &data) != 0))
	{
		DEBUG(0, ("tdb_store_vuid: tdb: %s\n", tdb_errorstr(tdb)));
		ret = False;
	}

	prs_free_data(&key);
	prs_free_data(&data);
	return ret;
}
