/*
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1998,2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1998,2000,
 *  Copyright (C) Paul Ashton                  1997-1998,2000,
 *  Copyright (C) Elrond                            2000,2003
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "debug.h"
#include "libsmb.h"
#include "nterr.h"
#include "doserr.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "others.h"


#define INIT_BASIC_VARS(comment) \
	prs_init(&buf, MARSHALL); \
	prs_init(&rbuf, UNMARSHALL); \
	\
	DEBUG(4, (comment))

#define CLEAN_AND_RET(ret) \
	rpccli_close_prs(&buf, &rbuf); \
	return (ret)



/****************************************************************************
do a SVC Open Policy
****************************************************************************/
BOOL svc_open_sc_man(const char *srv_name, const char *db_name,
		     uint32 des_access, POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_OPEN_SC_MAN q_c;
	SVC_R_OPEN_SC_MAN r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	struct cli_connection *con = NULL;

	if (!cli_connection_init(srv_name, PIPE_SVCCTL, &con))
	{
		return False;
	}

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Open SC_MAN\n");
	ZERO_STRUCT(r_c);

	make_svc_q_open_sc_man(&q_c, srv_name, db_name, des_access);

	/* turn parameters into data stream */
	if (!svc_io_q_open_sc_man("", &q_c, &buf, 0) ||
	    !rpc_con_pipe_req(con, SVC_OPEN_SC_MAN, &buf, &rbuf) ||
	    !svc_io_r_open_sc_man("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
	{
		DEBUG(1, ("SVC_OPEN_SC_MAN: %s\n",
			  get_nt_error_msg(status)));
		goto out;
	}

	/* ok, at last: we're happy. return the policy handle */
	*hnd = r_c.pol;
	if (!register_policy_hnd(get_global_hnd_cache(),
				 cli_con_sec_ctx(con),
				 hnd, des_access) ||
	    !set_policy_con(get_global_hnd_cache(), hnd,
			    con, cli_connection_unlink))
	{
		status = NT_STATUS_UNSUCCESSFUL;
	}

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Open Service
****************************************************************************/
BOOL svc_open_service(POLICY_HND *scm_hnd, const char *srv_name,
		      uint32 des_access, POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_OPEN_SERVICE q_c;
	SVC_R_OPEN_SERVICE r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || scm_hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Open Service\n");
	ZERO_STRUCT(r_c);

	make_svc_q_open_service(&q_c, scm_hnd, srv_name, des_access);

	/* turn parameters into data stream */
	if (!svc_io_q_open_service("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(scm_hnd, SVC_OPEN_SERVICE, &buf, &rbuf) ||
	    !svc_io_r_open_service("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
	{
		DEBUG(1, ("SVC_OPEN_SERVICE: %s\n",
			  get_nt_error_msg(status)));
		goto out;
	}

	/* ok, at last: we're happy. return the policy handle */
	*hnd = r_c.pol;
	if (!cli_pol_link(hnd, scm_hnd))
	{
		status = NT_STATUS_UNSUCCESSFUL;
	}

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Enumerate Services
****************************************************************************/
BOOL svc_enum_svcs(POLICY_HND *hnd,
		   uint32 services_type, uint32 services_state,
		   uint32 *buf_size, uint32 *resume_hnd,
		   uint32 *dos_error,
		   ENUM_SRVC_STATUS **svcs, uint32 *num_svcs)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_ENUM_SVCS_STATUS q_c;
	SVC_R_ENUM_SVCS_STATUS r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || buf_size == NULL || dos_error == NULL
	    || num_svcs == NULL)
	{
		return False;
	}

	INIT_BASIC_VARS("SVC Enum Services Status\n");
	ZERO_STRUCT(r_c);

	make_svc_q_enum_svcs_status(&q_c, hnd,
				    services_type, services_state,
				    *buf_size, *resume_hnd);

	/* turn parameters into data stream */
	if (!svc_io_q_enum_svcs_status("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_ENUM_SVCS_STATUS, &buf, &rbuf) ||
	    !svc_io_r_enum_svcs_status("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	if (r_c.dos_status != 0 && r_c.dos_status != ERRmoredata)
	{
		status = WIN32_TO_NTSTATUS(r_c.dos_status);
		DEBUG(1, ("SVC_ENUM_SVCS_STATUS: %s\n",
			  get_nt_error_msg(status)));
		goto out;
	}

	(*svcs) = r_c.svcs;
	(*num_svcs) = r_c.num_svcs;
	(*resume_hnd) = get_enum_hnd(&r_c.resume_hnd);
	(*buf_size) = r_c.more_buf_size;
	(*dos_error) = r_c.dos_status;
	status = NT_STATUS_OK;

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Stop Service
****************************************************************************/
BOOL svc_stop_service(POLICY_HND *hnd, uint32 unknown)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_STOP_SERVICE q_c;
	SVC_R_STOP_SERVICE r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Stop Service\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_stop_service(&q_c, hnd, unknown);

	/* turn parameters into data stream */
	if (!svc_io_q_stop_service("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_STOP_SERVICE, &buf, &rbuf) ||
	    !svc_io_r_stop_service("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_STOP_SERVICE: %s\n",
			  get_nt_error_msg(status)));

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Start Service
****************************************************************************/
BOOL svc_start_service(POLICY_HND *hnd, uint32 argc, char **argv)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_START_SERVICE q_c;
	SVC_R_START_SERVICE r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Start Service\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_start_service(&q_c, hnd, argc, argv);

	/* turn parameters into data stream */
	if (!svc_io_q_start_service("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_START_SERVICE, &buf, &rbuf) ||
	    !svc_io_r_start_service("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_START_SERVICE: %s\n",
			  get_nt_error_msg(status)));

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Query Service Config
****************************************************************************/
BOOL svc_query_svc_cfg(POLICY_HND *hnd,
		       QUERY_SERVICE_CONFIG *cfg, uint32 *buf_size)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_QUERY_SVC_CONFIG q_c;
	SVC_R_QUERY_SVC_CONFIG r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || buf_size == NULL)
		return False;

	INIT_BASIC_VARS("SVC Query Service Config\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_query_svc_config(&q_c, hnd, *buf_size);

	ZERO_STRUCTP(cfg);
	r_c.cfg = cfg;

	/* turn parameters into data stream */
	if (!svc_io_q_query_svc_config("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_QUERY_SVC_CONFIG, &buf, &rbuf) ||
	    !svc_io_r_query_svc_config("", &r_c, &rbuf, 0))
	{
		goto out;
	}
	if (r_c.buf_size == 0)
		goto out;

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_QUERY_SVC_CONFIG: %s\n",
			  get_nt_error_msg(status)));

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


/****************************************************************************
do a SVC Close
****************************************************************************/
BOOL svc_close(POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_CLOSE q_c;
	SVC_R_CLOSE r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Close\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_close(&q_c, hnd);

	/* turn parameters into data stream */
	if (!svc_io_q_close("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_CLOSE, &buf, &rbuf) ||
	    !svc_io_r_close("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_CLOSE: %s\n",
			  get_nt_error_msg(status)));

    out:
	close_policy_hnd(get_global_hnd_cache(), hnd);
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}

/****************************************************************************
do a SVC Change Service Config
****************************************************************************/
BOOL svc_change_svc_cfg(POLICY_HND *hnd,
			uint32 service_type, uint32 start_type,
			uint32 unknown_0, uint32 error_control,
			char *bin_path_name, char *load_order_grp,
			uint32 tag_id,
			char *dependencies, char *service_start_name,
			char *password, char *disp_name)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_CHANGE_SVC_CONFIG q_c;
	SVC_R_CHANGE_SVC_CONFIG r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Change Service Config\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_change_svc_config(&q_c, hnd,
				service_type, start_type,
	                        unknown_0, error_control,
				bin_path_name, load_order_grp,
				tag_id,
				dependencies, service_start_name,
				password, disp_name);

	/* turn parameters into data stream */
	if (!svc_io_q_change_svc_config("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_CHANGE_SVC_CONFIG, &buf, &rbuf) ||
	    !svc_io_r_change_svc_config("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_CHANGE_SVC_CONFIG: %s\n",
			  get_nt_error_msg(status)));

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


BOOL svc_lock(const POLICY_HND *scman_hnd, POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_LOCK q_c;
	SVC_R_LOCK r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (scman_hnd == NULL)
		return False;

	INIT_BASIC_VARS("SVC Unknown 3\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	q_c.scman_hnd = *scman_hnd;

	/* turn parameters into data stream */
	if (!svc_io_q_lock("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(scman_hnd, SVC_LOCK, &buf, &rbuf) ||
	    !svc_io_r_lock("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
	{
		DEBUG(1, ("SVC_LOCK: %s\n",
			  get_nt_error_msg(status)));
		goto out;
	}

	/* ok, at last: we're happy. return the policy handle */
	*hnd = r_c.hnd;
	if (!cli_pol_link(hnd, scman_hnd))
	{
		status = NT_STATUS_UNSUCCESSFUL;
		goto out;
	}
	policy_hnd_set_name(get_global_hnd_cache(),
			    hnd, "SVC_LOCK");

    out:
	CLEAN_AND_RET(NT_STATUS_IS_OK(status));
}


NTSTATUS svc_unlock(/* [in,out] */ POLICY_HND *hnd)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_CLOSE q_c;
	SVC_R_CLOSE r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS("SVC Unlock\n");
	ZERO_STRUCT(r_c);

	/* store the parameters */
	make_svc_q_close(&q_c, hnd);

	/* turn parameters into data stream */
	if (!svc_io_q_close("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_UNLOCK, &buf, &rbuf) ||
	    !svc_io_r_close("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_UNLOCK: %s\n",
			  get_nt_error_msg(status)));

    out:
	close_policy_hnd(get_global_hnd_cache(), hnd);
	CLEAN_AND_RET(status);
}


NTSTATUS svc_get_svc_sec(const POLICY_HND *hnd, uint32 sec_info,
			 /* [in,out] */ uint32 *buf_size,
			 SEC_DESC **sec_desc)
{
	prs_struct rbuf;
	prs_struct buf;
	SVC_Q_GET_SVC_SEC q_c;
	SVC_R_GET_SVC_SEC r_c;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (hnd == NULL || buf_size == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	INIT_BASIC_VARS("SVC Stop Service\n");
	ZERO_STRUCT(r_c);
	if(sec_desc)
		*sec_desc = NULL;

	/* store the parameters */
	q_c.hnd = *hnd;
	q_c.sec_info = sec_info;
	q_c.buf_size = *buf_size;

	/* turn parameters into data stream */
	if (!svc_io_q_get_svc_sec("", &q_c, &buf, 0) ||
	    !rpc_hnd_pipe_req(hnd, SVC_GET_SVC_SEC, &buf, &rbuf) ||
	    !svc_io_r_get_svc_sec("", &r_c, &rbuf, 0))
	{
		goto out;
	}

	status = WIN32_TO_NTSTATUS(r_c.status);
	if (!NT_STATUS_IS_OK(status))
		DEBUG(1, ("SVC_GET_SVC_SEC: %s\n",
			  get_nt_error_msg(status)));

	*buf_size = r_c.buf_size;
	if (sec_desc)
	{
		*sec_desc = r_c.sd;
	}
	else if (r_c.sd)
	{
		free_sec_desc(r_c.sd);
		free(r_c.sd);
	}

    out:
	CLEAN_AND_RET(status);
}
