/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1997,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1997,
 *  Copyright (C) Paul Ashton                       1997,
 *  Copyright (C) Lars Kneschke                     2000.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "winerror.h"
#include "others.h"


#define RPC_SERVER_STUB_DECLARE(name) \
	static BOOL name(PRS_HANDLE *hnd, prs_struct *data, prs_struct *rdata)
#define DO_UNMARSHALL(fn) \
	(fn("", &q_u, data, 0))
#define DO_MARSHALL(fn) \
	(fn("", &r_u, rdata, 0))



RPC_SERVER_STUB_DECLARE(api_reg_close)
{
	REG_Q_CLOSE q_u;
	REG_R_CLOSE r_u;
	ZERO_STRUCT(q_u);
	ZERO_STRUCT(r_u);

	if (!DO_UNMARSHALL(reg_io_q_close))
		return False;

	r_u.pol = q_u.pol;

	/* construct reply.  always indicate success */
	r_u.status = _reg_close(&r_u.pol);

	return DO_MARSHALL(reg_io_r_close);
}


RPC_SERVER_STUB_DECLARE(api_reg_open_hklm)
{
	REG_Q_OPEN_HKXX q_u;
	REG_R_OPEN_HKXX r_u;
	ZERO_STRUCT(q_u);
	ZERO_STRUCT(r_u);

	if (!DO_UNMARSHALL(reg_io_q_open_hkXX))
		return False;

	r_u.status = _reg_open_hklm(hnd, &r_u.pol, q_u.level);

	return DO_MARSHALL(reg_io_r_open_hkXX);
}


RPC_SERVER_STUB_DECLARE(api_reg_open_entry)
{
	REG_Q_OPEN_ENTRY q_u;
	REG_R_OPEN_ENTRY r_u;
	uint32 status;
	POLICY_HND entry_pol;

	ZERO_STRUCT(q_u);
	ZERO_STRUCT(r_u);

	if (!DO_UNMARSHALL(reg_io_q_open_entry))
		return False;

	/* construct reply. */
	status = _reg_open_entry(&q_u.pol, &q_u.uni_name, q_u.unknown_0,
				 q_u.access_mask, &entry_pol);

	make_reg_r_open_entry(&r_u, &entry_pol, status);

	return DO_MARSHALL(reg_io_r_open_entry);
}


RPC_SERVER_STUB_DECLARE(api_reg_query_key)
{
	REG_Q_QUERY_KEY q_u;
	REG_R_QUERY_KEY r_u;

	ZERO_STRUCT(q_u);
	ZERO_STRUCT(r_u);

	if (!DO_UNMARSHALL(reg_io_q_query_key))
		return False;

	/*
	 * should do the thing, but we're lazy...
	 * And leave the ZERO_STRUCT from above ;-)
	 */
	r_u.hdr_class.buffer = 1;

	return DO_MARSHALL(reg_io_r_query_key);
}


RPC_SERVER_STUB_DECLARE(api_reg_unknown_1a)
{
	REG_Q_UNK_1A q_u;
	REG_R_UNK_1A r_u;

	ZERO_STRUCT(q_u);
	ZERO_STRUCT(r_u);

	if (!DO_UNMARSHALL(reg_io_q_unk_1a))
		return False;

	/*
	 * should do the thing, but we're lazy...
	 * And leave the ZERO_STRUCT from above ;-)
	 */
	r_u.unknown = 0x05000000;

	return DO_MARSHALL(reg_io_r_unk_1a);
}


RPC_SERVER_STUB_DECLARE(api_reg_info)
{
	REG_R_INFO r_u;
	REG_Q_INFO q_u;
	BOOL ret;
	BUFFER2 buf;

	uint32 status;
	uint32 type = 0xcafeface;

	ZERO_STRUCT(r_u);
	ZERO_STRUCT(q_u);
	ZERO_STRUCT(buf);

	if (!DO_UNMARSHALL(reg_io_q_info))
		return False;

	/* construct reply.  always indicate success */
	status = _reg_info(&q_u.pol, &buf, &type);

	make_reg_r_info(q_u.ptr_buf, &r_u, &type, &buf, status);

	ret = DO_MARSHALL(reg_io_r_info);

	misc_buffer2_free(&buf);

	return ret;
}


RPC_SERVER_STUB_DECLARE(api_reg_get_key_sec)
{
	REG_Q_GET_KEY_SEC q_u;
	REG_R_GET_KEY_SEC r_u;
	SEC_DESC_BUF sd_buf;

	ZERO_STRUCT(r_u);
	ZERO_STRUCT(q_u);

	if (!DO_UNMARSHALL(reg_io_q_get_key_sec))
		return False;

	if (q_u.ptr)
		return False;

	ZERO_STRUCT(sd_buf);
	r_u.ptr = 1;
	r_u.hdr_sec.buf_max_len = 0x144;
	r_u.data = &sd_buf;
	sd_buf.max_len = 0x144;

	return DO_MARSHALL(reg_io_r_get_key_sec);
}


RPC_SERVER_STUB_DECLARE(api_reg_shutdown)
{
	REG_Q_SHUTDOWN q_u;
	REG_R_SHUTDOWN r_u;

	ZERO_STRUCT(r_u);
	ZERO_STRUCT(q_u);

	if (!DO_UNMARSHALL(reg_io_q_shutdown))
		return False;

	DEBUG(2, ("remote shutdown: (0x%x,0x%x,0x%x) timeout: %d, flags: 0x%x\n",
		  q_u.ptr_0, q_u.ptr_1, q_u.ptr_2, q_u.timeout, q_u.flags));

	r_u.status = ERROR_ACCESS_DENIED;

	return DO_MARSHALL(reg_io_r_shutdown);
}


RPC_SERVER_STUB_DECLARE(api_reg_abort_shutdown)
{
	REG_Q_ABORT_SHUTDOWN q_u;
	REG_R_ABORT_SHUTDOWN r_u;

	ZERO_STRUCT(r_u);
	ZERO_STRUCT(q_u);

	if (!DO_UNMARSHALL(reg_io_q_abort_shutdown))
		return False;

	r_u.status = ERROR_SUCCESS;

	return DO_MARSHALL(reg_io_r_abort_shutdown);
}


/*******************************************************************
 array of \PIPE\reg operations
 ********************************************************************/
static const struct api_struct api_reg_cmds[] = {
	{"REG_OPEN_HKLM", REG_OPEN_HKLM, api_reg_open_hklm},
	{"REG_OPEN_HKU", REG_OPEN_HKU, api_reg_open_hklm},
	{"REG_OPEN_HKPD", REG_OPEN_HKPD, NULL},
	{"REG_QUERY_KEY", REG_QUERY_KEY, api_reg_query_key},
	{"REG_UNK_1A", REG_UNK_1A, api_reg_unknown_1a},
	{"REG_CLOSE", REG_CLOSE, api_reg_close},
	{"REG_OPEN_ENTRY", REG_OPEN_ENTRY, api_reg_open_entry},
	{"REG_INFO", REG_INFO, api_reg_info},
	{"REG_GET_KEY_SEC", REG_GET_KEY_SEC, api_reg_get_key_sec},
	{"REG_ENUM_KEY", REG_ENUM_KEY, NULL},
	{"REG_ENUM_VALUE", REG_ENUM_VALUE, NULL},
	{"REG_CREATE_KEY", REG_CREATE_KEY, NULL},
	{"REG_SHUTDOWN", REG_SHUTDOWN, api_reg_shutdown},
	{"REG_ABORT_SHUTDOWN", REG_ABORT_SHUTDOWN, api_reg_abort_shutdown},
	{NULL, 0, NULL}
};

/*******************************************************************
 receives a reg pipe and responds.
 ********************************************************************/
BOOL api_reg_rpc(rpcsrv_struct * p)
{
	return api_rpcTNP(p, "api_reg_rpc", api_reg_cmds);
}
