/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Password and authentication handling
   Copyright (C) Andrew Tridgell 1992-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#ifdef HAVE_NETGROUP_H
#include <netgroup.h>
#endif
#include "libsamba.h"
#include "rpc_client.h"
#include "rpc_parse.h"
#include "nterr.h"
#include "smbd.h"

#undef  DBGC_CLASS
#define DBGC_CLASS DBGC_AUTH


/* users from session setup */
static pstring session_users = "";

extern pstring global_myname;


/****************************************************************************
add a name to the session users list
****************************************************************************/
void add_session_user(char *user)
{
  fstring suser;
  fstrcpy(suser, user);

  if (!Get_Pwnam(suser, True)) return;

	if (suser && *suser && !in_list(suser, session_users, False))
	{
		if (strlen(suser) + strlen(session_users) + 2 >= sizeof(pstring))
			DEBUG(1, ("Too many session users??\n"));
		else
		{
			pstrcat(session_users, " ");
			pstrcat(session_users, suser);
		}
	}
}

/****************************************************************************
This is a shortcut version of check_domain_security()
(in msrpc_netlogon.c), that always asks the local netlogond.
The local netlogond can then try to resolve this request remotely,
if it wants to.
****************************************************************************/
static uint32 ask_local_netlogond(const char *orig_user, const char *domain,
				  const uchar * challenge,
				  const char *smb_apasswd, int smb_apasslen,
				  const char *smb_ntpasswd, int smb_ntpasslen,
				  NET_USER_INFO_3 * info3)
{
	const char *acct_name;
	uint16 acct_type;

	/*
	 * This comes from check_domain_security().
	 * Is it realy needed?
	 * Elrond
	 */
	if (lp_security() == SEC_SHARE || lp_security() == SEC_SERVER)
	{
		return NT_STATUS_ACCESS_DENIED;
	}

	if (domain == NULL)
	{
		domain = "";
	}

	acct_name = global_myname;
	acct_type = SEC_CHAN_WKSTA;

	DEBUG(10, ("ask_local_netlogond: Checking for %s\\%s; %s(%d)\n",
		   domain, orig_user, acct_name, acct_type));

	return domain_client_validate("\\\\.", orig_user, domain,
				      acct_name, acct_type,
				      challenge,
				      smb_apasswd, smb_apasslen,
				      smb_ntpasswd, smb_ntpasslen, info3);
}

/****************************************************************************
check if a username/password pair is OK either via the system password
database or the encrypted SMB password database.
return NT_STATUS.
****************************************************************************/
uint32 password_ok(const char *orig_user, const char *domain,
		 const char *smb_apasswd, int smb_apasslen,
		 const char *smb_ntpasswd, int smb_ntpasslen,
		 struct passwd *pwd, NET_USER_INFO_3 * info3)
{
	uchar last_chal[8];
	BOOL cleartext = smb_apasslen != 24 && smb_ntpasslen == 0;
	uchar *chal = NULL;

	if (info3 == NULL)
	{
		DEBUG(0, ("password_ok: no NET_USER_INFO_3 parameter!\n"));
		return NT_STATUS_INVALID_PARAMETER;
	}

	DEBUG(5, ("password_ok(user=%s, domain=%s), cleartext=%s\n",
		  orig_user, domain, (cleartext ? "True" : "False")));

	/*
	 * SMB password check
	 */

	if ((smb_apasslen != 0) ||
	    (lp_encrypted_passwords() && smb_apasslen == 0 &&
	     lp_null_passwords()))
	{
		DEBUG(10, ("password_ok: check SMB auth\n"));

		/* check security = user / domain */
		if ((!cleartext) && last_challenge(last_chal))
		{
			chal = last_chal;
		}
		DEBUG(5, ("password_ok: chal=%s\n",
			  (chal ? "non-NULL" : "NULL")));
		if (cleartext || chal)
		{
			return ask_local_netlogond(orig_user, domain,
						chal,
						smb_apasswd, smb_apasslen,
						smb_ntpasswd, smb_ntpasslen,
						info3);
		}
	}

	DEBUG(5, ("password_ok: Wrong password\n"));

	return NT_STATUS_WRONG_PASSWORD;
}


/****************************************************************************
validate a group username entry. Return the username or NULL
****************************************************************************/
static char *validate_group(char *group, char *password, int pwlen, int snum,
		NET_USER_INFO_3 *info3)
{
#ifdef HAVE_NETGROUP
	char *host, *user, *domain;
#endif
#ifdef HAVE_GETGRENT
	struct group *gptr;
	pstring member_list;
	char *member;
	size_t copied_len = 0;
	int i;
#endif

#ifdef HAVE_NETGROUP

	setnetgrent(group);
	while (getnetgrent(&host, &user, &domain))
	{
		if (user)
		{
			if (user_ok(user, snum) &&
			   (password_ok(user, NULL, password, pwlen,
					NULL, 0, NULL, info3) == NT_STATUS_OK))
			{
				endnetgrent();
				return (user);
			}
		}
	}
	endnetgrent();
#endif

#ifdef HAVE_GETGRENT

	setgrent();
	while ((gptr = (struct group *)getgrent()))
	{
		if (strequal(gptr->gr_name, group))
			break;
	}

	/*
	 * As user_ok can recurse doing a getgrent(), we must
	 * copy the member list into a pstring on the stack before
	 * use. Bug pointed out by leon@eatworms.swmed.edu.
	 */

	if (gptr == NULL)
	{
		endgrent();
		return NULL;
	}

	*member_list = '\0';
	member = member_list;

	for (i = 0; gptr->gr_mem && gptr->gr_mem[i]; i++) {
		size_t member_len = strlen(gptr->gr_mem[i]) + 1;
		if (copied_len + member_len < sizeof(pstring)) {

			DEBUG(10, ("validate_group: = gr_mem = %s\n", gptr->gr_mem[i]));

			safe_strcpy(member, gptr->gr_mem[i], sizeof(pstring) - copied_len - 1);
			copied_len += member_len;
			member += copied_len;
		} else {
			*member = '\0';
		}
	}

	endgrent();

	member = member_list;
	while (*member) {
		static fstring name;
		fstrcpy(name, member);
		if (user_ok(name, snum) &&
		   (password_ok(name, NULL, password, pwlen,
				NULL, 0, NULL, info3) == NT_STATUS_OK))
		{
			endgrent();
			return (&name[0]);
		}

		DEBUG(10, ("validate_group = member = %s\n", member));

		member += strlen(member) + 1;
	}
#endif
	return (NULL);
}

/****************************************************************************
 Check for authority to login to a service with a given username/password.
****************************************************************************/

BOOL authorise_login(int snum, char *user, char *domain,
		     char *password, int pwlen,
		     BOOL *guest, BOOL *force, uint16 vuid)
{
	uint32 ok = NT_STATUS_WRONG_PASSWORD;
	user_struct *vuser = get_public_user_struct(vuid);

	DEBUG(3, ("authorise_login: TODO. split function, it's 6 levels!\n"));

#if DEBUG_PASSWORD
	DEBUG(100, ("authorise_login: checking authorisation on user=%s pass=%s\n",
		    user, password));
#endif

	*guest = False;

	if (GUEST_ONLY(snum))
		*force = True;

	if (!GUEST_ONLY(snum) && (lp_security() > SEC_SHARE)) {

		/*
		 * We should just use the given vuid from a sessionsetup_and_X.
		 */

		if (!vuser) {
			DEBUG(1,("authorise_login: refusing user %s with no session setup\n",
					user));
			return False;
		}

		if (!vuser->guest && user_ok(vuser->unix_name,snum)) {
			fstrcpy(user,vuser->unix_name);
			*guest = False;
			DEBUG(3,("authorise_login: ACCEPTED: validated uid ok as non-guest \
(user=%s)\n", user));
			ok = NT_STATUS_NOPROBLEMO;
		}
	}
 
	/* there are several possibilities:
	   1) login as the given user with given password
	   2) login as a previously registered username with the given password
	   3) login as a session list username with the given password
	   4) login as a previously validated user/password pair
	   5) login as the "user =" user with given password
	   6) login as the "user =" user with no password (guest connection)
	   7) login as guest user with no password

	   if the service is guest_only then steps 1 to 5 are skipped
	 */

	if (!(ok == NT_STATUS_OK)
	    && !(GUEST_ONLY(snum) && GUEST_OK(snum))) {

		/* check the given username and password */
		if (!(ok == NT_STATUS_OK) && (*user) && user_ok(user, snum)) {
			DEBUG(5, ("authorise_login: Trying given %s\\%s and given password\n",
				  domain, user));

			ok = password_ok(user, domain, password, pwlen,
					 NULL, 0, NULL, &vuser->usr);
			if (ok == NT_STATUS_OK)
				DEBUG(3, ("authorise_login: ACCEPTED: given username (%s) password ok\n",
					  user));
		}

		/* check for a previously registered guest username */
		if (!(ok == NT_STATUS_OK) && (vuser != 0) && vuser->guest) {
			ok = password_ok(vuser->unix_name, domain, password, pwlen,
								NULL, 0, NULL, &vuser->usr);
			if (user_ok(vuser->unix_name, snum) &&
			   (ok == NT_STATUS_OK))
			{
				fstrcpy(user, vuser->unix_name);
				vuser->guest = False;
				DEBUG(3, ("authorise_login: ACCEPTED: given password with registered user %s\n", user));
			}
		}

		/* now check the list of session users */
		if (!(ok == NT_STATUS_OK))
		{
			char *auser;
			char *user_list = strdup(session_users);

			DEBUG(5, ("authorise_login: checking session list users [%s]\n",
				  (user_list ? user_list : "(NULL)")));

			if (!user_list)
			{
				vuid_free_user_struct(vuser);
				return False;
			}

			for (auser = strtok(user_list, LIST_SEP);
			     !(ok == NT_STATUS_OK) && auser;
			     auser = strtok(NULL, LIST_SEP))
			{
				fstring user2;
				fstrcpy(user2, auser);
				if (!user_ok(user2, snum))
					continue;

				ok = password_ok(user2, domain, password, pwlen,
									NULL, 0, NULL, &vuser->usr);
				if (ok == NT_STATUS_OK)
				{
					fstrcpy(user, user2);
					DEBUG(3, ("ACCEPTED: session list username and given password ok\n"));
				}
			}

			free(user_list);
		}

		/* check for a previously validated username/password pair */
		if (!(ok == NT_STATUS_OK) && (lp_security() > SEC_SHARE)
		    && (vuser != 0) && !vuser->guest
		    && user_ok(vuser->unix_name, snum))
		{
			fstrcpy(user, vuser->unix_name);
			*guest = False;
			DEBUG(3, ("authorise_login: ACCEPTED: validated uid (%s) as non-guest\n",
				  user));
			ok = NT_STATUS_OK;
		}

		/* check for a rhosts entry */
		if (!(ok == NT_STATUS_OK) && user_ok(user, snum) && check_hosts_equiv(user)) {
			ok = NT_STATUS_OK;
			DEBUG(3, ("ACCEPTED: hosts equiv or rhosts entry\n"));
		}

		/* check the user= fields and the given password */
		if (!(ok == NT_STATUS_OK) && lp_username(snum)) {
			char *auser;
			pstring user_list;
			pstrcpy(user_list, lp_username(snum));

			pstring_sub(user_list, "%S", lp_servicename(snum));

			for (auser = strtok(user_list, LIST_SEP);
			     auser && !(ok == NT_STATUS_OK);
			     auser = strtok(NULL, LIST_SEP))
			{
				if (*auser == '@')
				{
					auser = validate_group(auser + 1,
							       password,
							       pwlen, snum,
							       &vuser->usr);
					if (auser)
					{
						ok = NT_STATUS_OK;
						fstrcpy(user, auser);
						DEBUG(3, ("ACCEPTED: group username and given password ok\n"));
					}
				}
				else
				{
					fstring user2;
					fstrcpy(user2, auser);
					ok = password_ok(user2, domain, password, pwlen, NULL, 0,
									 NULL, &vuser->usr);
					if (user_ok(user2, snum) &&
					   (ok == NT_STATUS_OK))
					{
						fstrcpy(user, user2);
						DEBUG(3, ("ACCEPTED: user list username and given password ok\n"));
					}
				}
			}
		}

		if (vuser != NULL)
		{
			tdb_store_vuid(vuid, vuser, True);
		}

	}			/* not guest only */

	/* check for a normal guest connection */
	if (!(ok == NT_STATUS_OK) && GUEST_OK(snum))
	{
		fstring guestname;
		fstrcpy(guestname, lp_guestaccount(snum));
		if (Get_Pwnam(guestname, True))
		{
			fstrcpy(user, guestname);
			ok = NT_STATUS_OK;
			DEBUG(3, ("authorise_login: ACCEPTED: guest account and guest ok (%s)\n",
				  user));
		}
		else
		{
			DEBUG(0, ("authorise_login: Invalid guest account %s??\n", guestname));
		}
		*guest = True;
	}

	if ((ok == NT_STATUS_OK) && !user_ok(user, snum))
	{
		DEBUG(0, ("authorise_login: rejected invalid user %s\n", user));
		ok = NT_STATUS_NO_SUCH_USER;
	}

	vuid_free_user_struct(vuser);
	return ((ok == NT_STATUS_OK));
}
